/*
 * Soft66 Radio Control for Linux
 * Copyright (C) 2010 Thomas Horsten <thomas@horsten.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "soft66.h"
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#define DEBUG 0
#if DEBUG
#define dprintf(args...) printf(args)
#else
#define dprintf(args...)
#endif

static int vid_from_env(void)
{
	const char *s;
	s = getenv(SOFT66_VID_ENV);
	return s ? strtoul(s, NULL, 0) : 0;
}

static int pid_from_env(void)
{
	const char *s;
	s = getenv(SOFT66_PID_ENV);
	return s ? strtoul(s, NULL, 0) : 0;
}

/**
 * Enumerate Soft66 devices found on the system and optionally call
 * a callback function with the serial number, manufacturer and description
 * for each (or NULL where these are not set).
 *
 * Callback function gets called with these arguments:
 *   cb(error, serial, manufacturer, description)
 *   error will be set to errno if getting the device info failed,
 *   usually EPERM.
 * 
 * @arg vid      Vendor ID (or 0 for default)
 * @arg pid      Product ID (or 0 for default)
 * @arg callback Callback function or NULL
 * 
 * Returns the number of devices found, or -1 if an error occurred.
 */
int soft66_list_devices(int vid, int pid, void(*callback)(const struct usb_device *, int, const char *, const char *, const char *))
{
	int r;
	struct ftdi_device_list *list, *e;
	struct ftdi_context *ftdi;
	char man[256];
	char des[256];
	char ser[256];
	int n = 0;

	ftdi = ftdi_new();
	if (!ftdi)
		return -1;

	if (vid == 0)
		vid = vid_from_env();
	if (vid == 0)
		vid = SOFT66_VID_DEFAULT;
	if (pid == 0)
		pid = pid_from_env();
	if (pid == 0)
		pid = SOFT66_PID_DEFAULT;
	
	r = ftdi_usb_find_all(ftdi, &list, vid, pid);
	if (r < 0) {
		n = -1;
		goto out;
	}
	for (e=list; e; e=list->next) {
		n++;
		r = ftdi_usb_get_strings(ftdi, e->dev,
					 man, sizeof(man)-1,
					 des, sizeof(des)-1,
					 ser, sizeof(ser)-1);
		if (callback) {
			if (r < 0) {
				strcpy(man, "Unknown");
				strcpy(ser, "Unknown");
				strcpy(des, "Unknown");
			}
			callback(e->dev, (r<0) ? errno : 0, ser, man, des);
		}
	}

out:
	if (list)
		ftdi_list_free(&list);
	if (ftdi) {
		ftdi_deinit(ftdi);
		ftdi_free(ftdi);
	}
	return n;

}

/**
 * Allocate and initialize a soft66_context
 *
 * Returns NULL on failure or the created context, which can later
 * be released with soft66_free().
 */
struct soft66_context *soft66_new(void)
{
	struct soft66_context *ctx;
	ctx = malloc(sizeof(struct soft66_context));
	if (!ctx)
		return NULL;
	memset(ctx, 0, sizeof(*ctx));
	ctx->ftdi = ftdi_new();
	if (!ctx->ftdi) {
		free(ctx);
		return NULL;
	}
	return ctx;
}

/**
 * Open Soft66 device. If vid and pid are specified as 0, the
 * values of the SOFT66_VID and SOFT66_PID environment variables
 * are used, and if they are also not set then the default values are
 * used. If serial is specified or if SOFT66_SERIAL is set in the
 * environment, then only a device with that serial number will be
 * opened. If it is NULL and not specified in the environment, the
 * first device matching vid and pid will be opened.
 *
 * @arg ctx    Context created by soft66_new()
 * @arg vid    Vendor ID (or 0 for default)
 * @arg pid    Product ID (or 0 for default)
 * @arg serial Serial number or NULL
 */
int soft66_open_device(struct soft66_context *ctx, 
			  int vid, int pid, const char *serial)
{
	int r;
	struct ftdi_device_list *list, *e;
	char ser[256];
	if (!serial)
		serial = getenv(SOFT66_SERIAL_ENV);
	if (vid == 0)
		vid = vid_from_env();
	if (vid == 0)
		vid = SOFT66_VID_DEFAULT;
	if (pid == 0)
		pid = pid_from_env();
	if (pid == 0)
		pid = SOFT66_PID_DEFAULT;
	
	r = ftdi_usb_find_all(ctx->ftdi, &list, vid, pid);
	if (r < 0)
		return r;
	if (!list) {
		errno = ENODEV;
		return -1;
	}
	r = 0;
	for (e=list; e; e=list->next) {
		r = ftdi_usb_get_strings(ctx->ftdi, e->dev,
					 NULL, 0,
					 NULL, 0,
					 ser, sizeof(ser)-1);
		if (r < 0) {
			continue;
		}
		if (serial && strcmp(ser, serial)) {
			r = -1;
			errno = ENOENT;
			continue;
		}
		if (!serial && e->next) {
			/* More than one device, we need to specify the
			 * serial number one way or another */
			r = -1;
			errno = EINVAL;
			break;
		}
		r = ftdi_usb_open_dev(ctx->ftdi, e->dev);
		ctx->model = SOFT66_MODEL_SOFT66ADD;
		ctx->freq_min = 1000;
		ctx->freq_max = 70000000;
		break;
	}
	ftdi_list_free(&list);
	return r;
}

int soft66_close_device(struct soft66_context *ctx)
{
	if (!ctx)
		return;
	if (ctx->ftdi) {
		ftdi_deinit(ctx->ftdi);
		ftdi_free(ctx->ftdi);
		ctx->ftdi = NULL;
	}
}

void soft66_free(struct soft66_context *ctx)
{
	if (!ctx)
		return;
	soft66_close_device(ctx);
	free(ctx);
}

static int write_16(unsigned char *buf, unsigned short val)
{
	int c, p = 0;
	int bit;
	for (c=0; c<4; c++)
		buf[p++] = 0x5; /* CLOCK + FSYNC high */
	buf[p++] = 0x1; /* FSYNC low */
	dprintf("w: %04x: ", val);
	for (c=0; c<16; c++) {
		bit = (val >> (15-c)) & 1;
		dprintf("%d ", bit);
		buf[p++] = (bit << 1) | 0x1; /* SCLK high */
		buf[p++] = (bit << 1); /* SCLK low */
	}
	buf[p++] = 0x5;
	dprintf("\n");
	return p;
}

int soft66_set_frequency(struct soft66_context *ctx, int freq, int noreset)
{
	unsigned char buf[2048];
	unsigned char *p;
	int len;
	int c;
	unsigned char filter_bits;
	unsigned int relfreq, ddefreq;
	unsigned short fr_lo, fr_hi;
	unsigned short ctrlreg;

	int r = 0;
	r = ftdi_set_bitmode(ctx->ftdi, 0xff, BITMODE_BITBANG);
	if (r<0) {
		errno = EIO;
		return -1;
	}
	r = ftdi_set_baudrate(ctx->ftdi, 30000);
	if (r<0) {
		errno = EIO;
		return -1;
	}

	/* Prepare buffer to write in bitbang mode. */
	if (freq < ctx->freq_min) {
		errno = EINVAL;
		return -1;
	}
	if (freq > ctx->freq_max) {
		errno = EINVAL;
		return -1;
	}

	switch (ctx->model) {
	case SOFT66_MODEL_SOFT66ADD:
		ddefreq = freq >> 1;
		break;
	default:
		ddefreq = freq;
	}

	/* Frequency as fraction of MCLK, normalized to 28 bits */
	relfreq = (unsigned int) (ddefreq / 75000000.0 * (double)(1 << 28));
	fr_hi = 0;
	fr_lo = 0;
	fr_lo = relfreq & 0x3fff;
	fr_hi = ( relfreq  >> 14 ) & 0x3fff;
	dprintf("ddefreq: %d relfreq: %d (0x%08x) - hi: 0x%04x lo: 0x%04x\n", ddefreq, relfreq, relfreq, fr_hi, fr_lo);
	p = buf;

	filter_bits = 0;
	if (freq < 1200000)
		filter_bits = 0x10;
	else if (freq < 5500000)
		filter_bits = 0x30;
	else if (freq < 10500000)
		filter_bits = 0x40;
	else if (freq < 18500000)
		filter_bits = 0x60;
	else if (freq < 30000000)
		filter_bits = 0x00;
	else
		filter_bits = 0x70;

	if ((freq < 1600000) || ( freq> 19000000))
		ctrlreg = 0x2028;
	else
		ctrlreg = 0x2038;

	if (!noreset)
        {
		/* Write the control register */
		p += write_16(p, ctrlreg | 0x0100);
	}

	/* Top two bits is the register address: 
	 * 1 for FREQ0
	 * 2 for FREQ1 
	 * 3 for PHASE
	 */
	/* FREQ0 */
	p += write_16(p, 0x4000 | fr_lo);
	p += write_16(p, 0x4000 | fr_hi);
	/* FREQ1 */
	p += write_16(p, 0x8000 | fr_lo);
	p += write_16(p, 0x8000 | fr_hi);
	if (!noreset) {
		/* PHASE register */
		p += write_16(p, 0xc000);
	}
	/* Write control register again with RESET disabled */
	p += write_16(p, ctrlreg);

	len = p-buf;

	/* Or in the filter control bits */
	for (c=0; c<len; c++)
		buf[c] |= filter_bits;
		

	dprintf("To write: %d bytes. Filter bits: 0x%02x\n", len, filter_bits);
	dprintf("Bits being written:\n");
	for (c=0; c<len; c++)
		if ((buf[c] & 1) == 0) /* Falling SCLK edge */
			dprintf("%d", (buf[c] & 0x2) >> 1);
	//for (c=0; c<len; c++)
	//	printf("%d: 0x%02x\n", c, buf[c]);
	//printf("\n");
	r = ftdi_write_data(ctx->ftdi, (unsigned char *)buf, len);
	return r;
}


int soft66_set_filterbits(struct soft66_context *ctx, unsigned char filterbits)
{
	int r = 0;
	r = ftdi_set_bitmode(ctx->ftdi, 0xff, BITMODE_BITBANG);
	if (r<0) {
		errno = EIO;
		return -1;
	}
	r = ftdi_set_baudrate(ctx->ftdi, 30000);
	if (r<0) {
		errno = EIO;
		return -1;
	}

	filterbits &= ~0x7;
	filterbits |= 0x5;
	r = ftdi_write_data(ctx->ftdi, (unsigned char *)&filterbits, 1);

	return r;
}

int soft66_get_filterbits(struct soft66_context *ctx, unsigned char *filterbits)
{
	int r = 0;
	r = ftdi_set_bitmode(ctx->ftdi, 0xff, BITMODE_BITBANG);
	if (r<0) {
		errno = EIO;
		return -1;
	}
	r = ftdi_set_baudrate(ctx->ftdi, 30000);
	if (r<0) {
		errno = EIO;
		return -1;
	}

	r = ftdi_read_data(ctx->ftdi, filterbits, 1);
	return r;
}

int soft66_sleep(struct soft66_context *ctx)
{
	int r = 0;
	int len;
	char buf[256];

	r = ftdi_set_bitmode(ctx->ftdi, 0xff, BITMODE_BITBANG);
	if (r<0) {
		errno = EIO;
		return -1;
	}
	r = ftdi_set_baudrate(ctx->ftdi, 30000);
	if (r<0) {
		errno = EIO;
		return -1;
	}

	len = write_16(buf, 0x21f8);
	r = ftdi_write_data(ctx->ftdi, (unsigned char *)buf, len);

	return r;
}
