/*
 * Soft66 Radio Control for Linux
 * Copyright (C) 2010 Thomas Horsten <thomas@horsten.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "soft66.h"
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#define _GNU_SOURCE
#include <getopt.h>

struct soft66_context *ctx = NULL;

#define die(args...) do { fprintf(stderr, ##args); exit(1); } while(0)

/* These defaults can be overridden by command line options */
static const char *serial = NULL;
static int show_list = 0;
static int show_info = 0;
static unsigned long tune_freq = 0;
static int retune = 0;
static int vid = 0;
static int pid = 0;
static int get_filterbits = 0;
static int set_filterbits = 0;
static int powerdown = 0;
static unsigned char filterbits = 0;


/* Show usage information */
static void show_help(const char *cmdname)
{
  printf("Usage:\n"
	 "%s [OPTIONS]... COMMAND [ARGS]...\n"
	 "       Controls the operation of the Soft66 series SDR radio receivers.\n"
	 "\n"
	 "Options:\n"
	 "  -l, --list           List attached Soft66 devices and exit.\n"
	 "  -b, --filterbits=B   Set the filterbits to the value specified (only).\n"
	 "  -i, --info           Query device and show status information.\n"
	 "  -s, --serial=SER     Operate on the device with serial number SER.\n"
	 "  -t, --tune=FREQ      Tune to FREQ Hz (postfix k or m for kHz or MHz, kHz is\n"
	 "                       assumed if a value below 70000 is given).\n"
	 "  -r, --retune=FREQ    Like --tune but do not reset the AD9834 first.\n"
	 "  -p, --powerdown      Put the receiver to sleep.\n"
	 "  -V, --vendorid=ID    Use device with specified vendor ID.\n"
	 "  -P, --productid=ID   Use device with specified product ID.\n"
	 "\n"
	 "Report bugs to <thomas@horsten.com>.\n", cmdname);
}

/* Parse options. Return index of first unconsumed argument. */
static int parse_arguments(int argc, char *argv[])
{
	while (1) 
	{
		static struct option long_options[] =
			{
				/* These options set a flag. */
				//{"verbose", d_argument,       &verbose_flag, 1},
				//{"pidfile",   no_argument,       &verbose_flag, 0},
				/* These options don't set a flag.
				   We distinguish them by their indices. */
				{"help",      no_argument,       0, 'h'},
				{"list",      no_argument,       0, 'l'},
				{"serial",    required_argument, 0, 's'},
				{"info",      no_argument,       0, 'i'},
				{"tune",      required_argument, 0, 't'},
				{"retune",    required_argument, 0, 'r'},
				{"powerdown", no_argument, 0, 'p'      },
				{"filterbits",required_argument, 0, 'b'},
				{"vendorid",  required_argument, 0, 'V'},
				{"productid", required_argument, 0, 'P'},
				{0, 0, 0, 0}
			};
		/* `getopt_long' stores the option index here. */
		int option_index = 0;
      
		int c = getopt_long (argc, argv, "hls:it:r:b:V:P:p",
				     long_options, &option_index);

		char * endptr;

		/* Detect the end of the options. */
		if (c == -1)
			break;
      
		switch (c)
		{
		case 'h':
			show_help(argv[0]);
			exit(0);
			break;

		case 'l':
			show_list = 1;
			break;

		case 's':
			serial=optarg;
			break;

		case 'p':
			powerdown=1;
			break;

		case 'r':
			retune = 1;
			/* fall through */
		case 't':
			tune_freq = strtoul(optarg, &endptr, 0);
			if (*endptr) {
				if (*endptr == 'k' || *endptr == 'K')
					tune_freq = tune_freq * 1000;
				else if (*endptr == 'm' || *endptr == 'M')
					tune_freq = tune_freq * 1000000;
				else die("Invalid frequency specified\n");

			} else {
				if (tune_freq < 70000) {
					tune_freq *= 1000;
				}
			}
			break;
	  
		case 'i':
			show_info = 1;
			break;
	  
		case 'V':
			vid = strtoul(optarg, &endptr, 0);
			if (*endptr) {
				die("Invalid VID value '%s'\n", optarg);
			}
			break;

		case 'P':
			pid = strtoul(optarg, &endptr, 0);
			if (*endptr) {
				die("Invalid PID value '%s'\n", optarg);
			}
			break;

		case 'b':
			set_filterbits = 1;
			filterbits = strtoul(optarg, &endptr, 0);
			if (*endptr) {
				die("Invalid filterbits value '%s'\n", optarg);
			}
			break;
	  	  
		case '?':
			/* `getopt_long' already printed an error message. */
			break;
	  
		default:
			fprintf(stderr, "Unexpected error\n");
			exit(1);
		}
	}
  
	/* Print any remaining command line arguments (not options). */
	if (optind < argc)
	{
		die("Unexpected non-option arguments on command line\n");
	}
	return optind;
}

static void list_cb(const struct usb_device *dev, int error, const char *ser, const char *man, const char *des)
{
	printf("Bus %s device %s:\n", dev->bus->dirname, dev->filename);
	if (error) {
		printf("  Device information incomplete: %s\n",
		       strerror(error));
		if (error == EPERM) {
			printf("  You do not have permission to access this device. See the documentation!\n");
		}
	}
	printf("  Manufacturer: %s\n"
	       "  Description:  %s\n"
	       "  Serial no.:   %s\n", man, des, ser);
}

static void show_device_info(void)
{
	int r;
	unsigned char filterbits;
	r = soft66_get_filterbits(ctx, &filterbits);
	if (r < 0) {
		die("Query device failed: %s\n", strerror(errno));
	}
	printf("Register setting: 0x%02x\n", filterbits);
}

int main(int argc, char **argv)
{
	int r;

	parse_arguments(argc, argv);

	if (show_list) {
		r = soft66_list_devices(vid, pid, list_cb);
		printf("%d device%s found\n", r, r==1?"":"s");
		exit(r == 0 ? 1 : 0);
	}

	if ((!show_info) && (!tune_freq) && (!set_filterbits) && (!powerdown))
		die("Nothing to do, try %s -h for help\n", argv[0]);

	ctx = soft66_new();
	if (!ctx)
		die("Couldn't create context: %s\n", strerror(errno));
	r = soft66_open_device(ctx, vid, pid, serial);
	if (r < 0) {
		die("Couldn't open Soft66 device: %s\n", strerror(errno));
	}

	if (show_info) {
		show_device_info();
	}

	if (tune_freq) {
		r = soft66_set_frequency(ctx, tune_freq, retune);
		if (r < 0) {
			die("Couldn't set frequency %.3fkHz: %s\n", tune_freq/1000.0, strerror(errno));
		}
		printf("Tuned to %.3fkHz\n", tune_freq/1000.0);
	}

	if (set_filterbits) {
		r = soft66_set_filterbits(ctx, filterbits);
		if (r < 0) {
			die("Couldn't set filterbits: %s\n", strerror(errno));
		}
		printf("Filterbits set to 0x%02x\n", filterbits);
	}

	if (powerdown) {
		r = soft66_sleep(ctx);
		if (r < 0) {
			die("Couldn't set sleep mode: %s\n", strerror(errno));
		}
		printf("Radio put to sleep\n", filterbits);
	}

	exit(0);
}
